/**************************************************************************
 * 
 * PROJECT: Task Control Management.
 *
 * (c) Copyright 1996 Reid Simmons.  All rights reserved.
 *
 * FILE: agenda.h
 *
 * ABSTRACT: Defines the task agenda, which is responsible for 
 *           sending signals, executing the actions of task tree nodes, and
 *           other TBD functions.
 *
 * EXPORTS:
 *
 * $Revision: 1.9 $
 * $Date: 1999/08/04 14:00:16 $
 * $Author: reids $
 *
 * REVISION HISTORY:
 *
 * $Log: agenda.h,v $
 * Revision 1.9  1999/08/04 14:00:16  reids
 * Distributed version of TCM.  Provides the ability for one process to
 *  add tasks, constraints, exception handlers, etc. with another process.
 *  The distributed version is created by compiling with -DDISTRIBUTED flag,
 *   which can be gotten by "gmaking" with DISTRIBUTED=1.
 *
 * Revision 1.8  1998/12/16 03:08:19  reids
 * Added support for "on termination" functions.
 *   Also enabled tca.h and tcm.h to co-exist (needed to change values
 *   of several constants).
 *
 * Revision 1.7  98/10/30  11:16:31  da0g
 * Added ExternalEventBypassHandler.
 * 
 * Revision 1.6  1998/04/21 12:45:53  reids
 * Added InvokeWhen/InvokeAfter functionality, to invoke a function when a
 *   particular event occurs.
 * Added the DelayUntilAfter/TerminateAtAfter functionality, to delay/terminate
 *   a node by waiting some msecs after a particular event has occurred.
 *
 * Revision 1.5  97/12/30  13:38:38  reids
 * When node is terminated, signal (and remove) any timed-event in the timer
 *   queue that is for that node, regardless of when it is supposed to fire.
 * 
 * Revision 1.4  97/12/30  12:24:45  reids
 * Added option to *not* wait for timeouts and external events.
 * 
 * Revision 1.3  97/12/29  17:06:14  reids
 * Version that has the basic functionality needed to support TDL.
 * 
 * Revision 1.2  97/12/04  17:50:04  reids
 * Another fairly stable version (except that monitors do not quite work)
 * 
 * Revision 1.1  97/11/21  14:06:23  reids
 * First release of TCM -- seems to be a stable version
 * 
 *
 **************************************************************************/

#ifndef INCagenda
#define INCagenda

#include "tcmBasics.h"
#include "queue.h"
#include "event.h"
#include "external.h"
#include "tplConstr.h"

#define TIMER_DATA ((void *)-1)

typedef BOOLEAN (*Timed_Event_Comp)(const Timed_Event &, const Timed_Event &);
extern BOOLEAN timerCompFn(const Timed_Event &, const Timed_Event &);
typedef priority_queue<Timed_Event, Timed_Event_Comp> Timer_Queue;

typedef queue<Pending_Event> Event_Queue;

typedef queue<Task_Tree_Ref> Task_Queue;

class Agenda
{
 public:
  Agenda() : _timerQueue(timerCompFn), _externalEvents() { }
  void queueEvent(const Pending_Event &event) { _eventQueue.push(event); };
  void queueTask(Task_Tree_Ref const &task) { _taskQueue.push(task); };
  void queueTimer(Timed_Event const &timedEvent) 
    { _timerQueue.push(timedEvent); };
  void removeTimer(Timed_Event const &timedEvent);

  void addExternalEvent (int sd, SD_CALLBACK_FN_TYPE callback, 
			 const void *clientData)
	  { _externalEvents.addExternalEvent(sd, callback, clientData); }
  void removeExternalEvent (int sd) { _externalEvents.removeExternalEvent(sd);}
  void setExternalEventBypassHandler (
			  EXTERNAL_EVENT_BYPASS_HANDLER_TYPE theBypassHandler )
		     { _externalEvents.setBypassHandler ( theBypassHandler ); }

  void clearQueues(Time_Point *timePoint, MSecs timeout,
		   BOOLEAN waitingAllowed);

  BOOLEAN taskQueued(Task_Tree_Ref const &task) const
    { return _taskQueue.member(task); }

 private:
  Event_Queue _eventQueue;
  Task_Queue  _taskQueue;
  Timer_Queue _timerQueue;
  External_Events _externalEvents;

  void _clearEventQueue (void);
  void _clearTimerQueue (void);
  BOOLEAN _popTaskQueue (void);
  BOOLEAN _waitForEvents (MSecs timeout);
  BOOLEAN _doneProcessing (Time_Point *timePoint, MSecs absoluteTime)
    { return ((timePoint && timePoint->isPast()) ||
	      (absoluteTime != INFINITE_TIME && 
	       absoluteTime <= timeInMsecs())); }
  struct timeval *_selectTimeout (MSecs maxTime);
};

void addTimedEvent (MSecs absoluteTime,
		    const Task_Tree_Ref nodeRef, Signal_Enum signal);

void addTimedEvent (MSecs absoluteTime,
		    SIMPLE_CALLBACK_FN_TYPE callbackFn, const void *data);

#endif /* INCagenda */
