/*****************************************************************************
 * 
 * PROJECT: Task Control Management.
 *
 * (c) Copyright 1996 Reid Simmons.  All rights reserved.
 *
 * FILE: tcm.cc
 *
 * ABSTRACT: Public include file for the Task Control Management library.
 *           Defines the TCM API.
 *
 * $Source: /afs/cs/project/TCA/Master/tcaV8/tcm/src/tcm.cc,v $ 
 * $Revision: 1.17 $
 * $Date: 1999/06/06 13:48:09 $
 * $Author: reids $
 *
 * REVISION HISTORY:
 *
 * $Log: tcm.cc,v $
 * Revision 1.17  1999/06/06 13:48:09  reids
 * Changes to TCM 1.3.0:
 *   New list data structures.
 *   New, generalized logging interface.
 *   Fixed firstChild to work if there are no children.
 *   Fixed code to deallocate a node.
 *   In certain circmumstances, don't send termination signal when task
 *     itself terminating.
 *
 * Revision 1.16  1999/03/02 01:31:13  da0g
 * Updated TCM_FailureData to use getNodeRefCountData
 *
// Revision 1.15  98/12/21  17:56:11  reids
// Added TCM_GetAction.
// 
 * Revision 1.14  1998/12/16 03:08:20  reids
 * Added support for "on termination" functions.
 *   Also enabled tca.h and tcm.h to co-exist (needed to change values
 *   of several constants).
 *
// Revision 1.13  98/12/14  15:20:08  da0g
// Added relativeTimeout to TCM_ProcessAgenda
// 
// Revision 1.12  98/10/30  11:16:32  da0g
// Added ExternalEventBypassHandler.
// 
 * Revision 1.11  1998/09/15 18:45:22  da0g
 * Enhanced exceptions to support multiple-name resolution and Ref_Count (automatically-destroyed) Data.
 *
 * Revision 1.10  1998/07/14 17:21:02  reids
 * Changed TCM_DelayUntilAfter to TCM_DelayForAfter, changed TCM_
 *   TerminateAtAfter to TCM_TerminateInAfter to TCM_DelayForAfter, and
 *   changed order of arguments.
 * Added TCM_ActivateInAfter.
 *
// Revision 1.9  98/04/21  12:46:10  reids
// Added InvokeWhen/InvokeAfter functionality, to invoke a function when a
//   particular event occurs.
// Added the DelayUntilAfter/TerminateAtAfter functionality, to delay/terminate
//   a node by waiting some msecs after a particular event has occurred.
// 
// Revision 1.8  98/03/06  13:12:45  reids
// Added a way to stop the compiler from complaining about unused function
//   arguments (apparently, g++ ignores the "#pragma unused" statement).
// 
// Revision 1.7  98/01/30  14:50:59  reids
// Updated to compile under gcc 2.7.2 and under Linux.
// Also, made STRING "const char *" and changed API to take const arguments,
//   where applicable.
// 
// Revision 1.6  97/12/30  12:29:02  reids
// Added option to *not* wait for timeouts and external events.
//   Added flag to indicate whether node raised an exception (ie, failed).
//   Added a "timer" monitor, which is like a polling monitor, except when
//     activated it just invokes a function (rather than adding a new node)
// 
// Revision 1.5  97/12/29  17:06:32  reids
// Version that has the basic functionality needed to support TDL.
// 
// Revision 1.4  97/12/22  16:53:03  reids
// Basically, added "data" field for CALLBACK_ACTION,
//  and started using "nodeData" for the activation data associated
//  with monitors, and the failure data associated with exceptions.
// 
// Revision 1.3  97/12/18  00:21:49  reids
// Changing ACTION_PTR to a handle, to aid in garbage collection.
// 
// Revision 1.2  97/12/04  17:50:21  reids
// Another fairly stable version (except that monitors do not quite work)
// 
// Revision 1.1  97/11/21  14:06:39  reids
// First release of TCM -- seems to be a stable version
// 
 *
 *****************************************************************************/

#include "tcmGlobal.h"
#include "taskTree.h"
#include "tplConstr.h"
#include "exception.h"
#include "monitor.h"
#include "tcm.h"

const TCM_Task_Tree_Ref NullRef;

#define RETURN_IF_INVALID_NODE(nodeRef, fnName) \
  if (!checkNodeValidity(nodeRef, fnName)) return TCM_Error;
#define RETURN_ON_INVALID_NODE(nodeRef, fnName, retVal) \
  if (!checkNodeValidity(nodeRef, fnName)) return retVal;
#define RETURN_IF_INVALID_TIME(timePoint, fnName) \
  if (!checkTimePointValidity(timePoint, fnName)) return TCM_Error;
#define RETURN_IF_NULL(value, fnName) \
  if (value == NULL) { \
    tcmWarning("%s: NULL argument\n", fnName); return TCM_Error; }

/*****************************************************************
 * 
 * Utility functions.
 *
 *****************************************************************/

static BOOLEAN checkNodeValidity (const TCM_Task_Tree_Ref &nodeRef,
				  STRING functionName)
{
  BOOLEAN valid = validRef(nodeRef);
  if (!valid) tcmWarning("%s: Reference to NULL node\n", functionName);
  return valid;
}

static BOOLEAN checkTimePointValidity (const TCM_Point &timePoint,
				       STRING functionName)
{
  BOOLEAN valid = timePoint.valid();
  if (!valid) tcmWarning("%s: Reference to invalid time point\n", functionName);
  return valid;
}

/*****************************************************************
 * 
 * Functions having to do with creating task tree nodes
 *   and defining handlers/callbacks.
 *
 *****************************************************************/

TCM_Task_Tree_Ref const &TCM_RootNode (void)
{
  return GET_TCM_GLOBAL(rootNode);
}

/**************************************************************************
 *
 * FUNCTION: TCM_Task_Tree_Ref TCM_AllocateGoalNode (STRING name)
 *
 * DESCRIPTION: Create a new goal node, with the given name
 *
 * NOTE: The node is not connected to the task tree, and is not runnable.
 *
 **************************************************************************/

TCM_Task_Tree_Ref TCM_AllocateGoalNode (STRING name)
{
  return new Task_Tree_Goal_Node(name);
}

/**************************************************************************
 *
 * FUNCTION: TCM_Task_Tree_Ref TCM_AllocateCommandNode (STRING name)
 *
 * DESCRIPTION: Create a new command node, with the given name
 *
 * NOTE: The node is not connected to the task tree, and is not runnable.
 *
 **************************************************************************/

TCM_Task_Tree_Ref TCM_AllocateCommandNode (STRING name)
{
  return new Task_Tree_Command_Node(name);
}

/**************************************************************************
 *
 * FUNCTION: TCM_Return_Type 
 *           TCM_DeallocateNode (const TCM_Task_Tree_Ref &nodeRef)
 *
 * DESCRIPTION: Destroy the node.
 *
 * NOTE: Only works if the node is not connected to the task tree,
 *       and has no children!  Otherwise, use "TCM_Terminate"
 *
 **************************************************************************/

TCM_Return_Type TCM_DeallocateNode (const TCM_Task_Tree_Ref &nodeRef)
{
  RETURN_IF_INVALID_NODE(nodeRef, "TCM_DeallocateNode");

  return (nodeRef->deallocate() ? TCM_Ok : TCM_Error);
}

/**************************************************************************
 *
 * FUNCTION: TCM_Return_Type TCM_SetAction (const TCM_Task_Tree_Ref &nodeRef, 
 *                                          const TCM_Action_Ref action);
 *
 * DESCRIPTION: Set the action (handler function) for the given node.
 *
 **************************************************************************/

TCM_Return_Type TCM_SetAction (const TCM_Task_Tree_Ref &nodeRef, 
			       const TCM_Action_Ref action)
{
  RETURN_IF_INVALID_NODE(nodeRef, "TCM_SetAction");

  nodeRef->setAction(action);
  return TCM_Ok;
}

/**************************************************************************
 *
 * FUNCTION: TCM_Return_Type TCM_GetAction (const TCM_Task_Tree_Ref &nodeRef, 
 *                                          const TCM_Action_Ref action);
 *
 * DESCRIPTION: Get the action (handler function) for the given node.
 *
 **************************************************************************/

TCM_Action_Ref TCM_GetAction (const TCM_Task_Tree_Ref &nodeRef)
{
  RETURN_ON_INVALID_NODE(nodeRef, "TCM_GetAction", NULL);

  return nodeRef->getAction();
}

/**************************************************************************
 *
 * FUNCTION: TCM_Return_Type TCM_SetData (const TCM_Task_Tree_Ref &nodeRef, 
 *                                        const void *data);
 *
 * DESCRIPTION: Set the data associated with the given node.
 *
 **************************************************************************/

TCM_Return_Type TCM_SetData (const TCM_Task_Tree_Ref &nodeRef,
			     const void *data)
{
  RETURN_IF_INVALID_NODE(nodeRef, "TCM_SetData");

  nodeRef->setNodeData(data);
  return TCM_Ok;
}

/**************************************************************************
 *
 * FUNCTION: TCM_InsertNode (const TCM_Task_Tree_Ref &parentRef,
 *			     const TCM_Task_Tree_Ref &childRef,
 *			     BOOLEAN isLastChild);
 *
 * DESCRIPTION: Add the childRef node as a child of the parentRef, and
 *              enable the childRef to be run.
 *
 **************************************************************************/

TCM_Return_Type TCM_InsertNode (const TCM_Task_Tree_Ref &parentRef,
				const TCM_Task_Tree_Ref &childRef,
				BOOLEAN isLastChild)
{
  RETURN_IF_INVALID_NODE(parentRef, "TCM_InsertNode");
  RETURN_IF_INVALID_NODE(childRef, "TCM_InsertNode");

  taskTreeInsert(*parentRef, *childRef, isLastChild);

  return TCM_Ok;
}

/****************************************************************
 *
 * The following four functions are shorthands for a sequence of
 *  allocate/setAction/insert calls.
 ****************************************************************/

TCM_Task_Tree_Ref TCM_CreateGoalNode (const TCM_Task_Tree_Ref &parentRef,
				      STRING name,
				      const TCM_Action_Ref &action,
				      BOOLEAN isLastChild)
{
  RETURN_ON_INVALID_NODE(parentRef, "TCM_CreateGoalNode", NullRef);

  TCM_Task_Tree_Ref nodeRef(new Task_Tree_Goal_Node(name, action));
  taskTreeInsert(*parentRef, *nodeRef, isLastChild);

  return nodeRef;
}

TCM_Task_Tree_Ref TCM_CreateGoalNode (const TCM_Task_Tree_Ref &parentRef,
				      STRING name, 
				      const void *callbackData,
				      CALLBACK_FN_TYPE callback,
				      BOOLEAN isLastChild)
{
  return TCM_CreateGoalNode(parentRef, name,
			    new Callback_Action(callback, callbackData),
			    isLastChild);
}

TCM_Task_Tree_Ref TCM_CreateCommandNode (const TCM_Task_Tree_Ref &parentRef,
					 STRING name,
					 const TCM_Action_Ref &action,
					 BOOLEAN isLastChild)
{
  RETURN_ON_INVALID_NODE(parentRef, "TCM_CreateCommandNode", NullRef);

  TCM_Task_Tree_Ref nodeRef(new Task_Tree_Command_Node(name, action));
  taskTreeInsert(*parentRef, *nodeRef, isLastChild);

  return nodeRef;
}

TCM_Task_Tree_Ref TCM_CreateCommandNode (const TCM_Task_Tree_Ref &parentRef,
					 STRING name, 
					 const void *callbackData,
					 CALLBACK_FN_TYPE callback,
					 BOOLEAN isLastChild)
{
  return TCM_CreateCommandNode(parentRef, name,
			       new Callback_Action(callback, callbackData),
			       isLastChild);
}

TCM_Return_Type TCM_Success (const TCM_Task_Tree_Ref &nodeRef)
{
  RETURN_IF_INVALID_NODE(nodeRef, "TCM_Success");

  nodeRef->doneHandling();
  return TCM_Ok;
}

TCM_Return_Type TCM_Failure (const TCM_Task_Tree_Ref &nodeRef, 
			     STRING failure, void *failureData)
{
  RETURN_IF_INVALID_NODE(nodeRef, "TCM_Failure");

  if (!failure || *failure == '\0') {
    tcmWarning("TCM_Failure: No failure specified\n");
    return TCM_Error;
  } else {
    raiseException(failure, nodeRef, failureData);
    return TCM_Ok;
  }
}

TCM_Return_Type TCM_Failure ( const TCM_Task_Tree_Ref & theNodeRef, 
			      TCM_StringMatcher         theStringMatcher,
			      const char              * theFailureId,
			      Ref_Count               * theFailureData )
{
  RETURN_IF_INVALID_NODE ( theNodeRef, "TCM_Failure" );

  if ( theStringMatcher == TCM_StringMatcher(NULL) )
  {
    tcmWarning ( "TCM_Failure: No failure string-matcher specified\n" );
    return TCM_Error;
  }

  else if (   ( theFailureId    == (const char *) NULL )
	   || ( theFailureId[0] == (const char  ) NULL ) )
  {
    tcmWarning ( "TCM_Failure: No failure-Id string specified.  "
		 "Using \"UNKNOWN\"\n");
    raiseException ( theStringMatcher, "UNKNOWN", theNodeRef, theFailureData );
    return TCM_Ok;
  }

  else
  {
    raiseException ( theStringMatcher, theFailureId,
		     theNodeRef,       theFailureData );
    return TCM_Ok;
  }
}

/**************************************************************************
 *
 * Functions to map between a task tree node and its id number
 *
 **************************************************************************/
 
/**************************************************************************
 *
 * FUNCTION: int TCM_Remember (const TCM_Task_Tree_Ref &nodeRef)
 *
 * DESCRIPTION: Memoize the node, so that it can be recalled
 *
 **************************************************************************/
int TCM_Remember (const TCM_Task_Tree_Ref &nodeRef)
{
  RETURN_ON_INVALID_NODE(nodeRef, "TCM_Remember", -1);

  return nodeRef->remember();
}

/**************************************************************************
 *
 * FUNCTION: TCM_Task_Tree_Ref TCM_Recall (int refID)
 *
 * DESCRIPTION: Find the task tree node associated with the give id number
 *
 **************************************************************************/
TCM_Task_Tree_Ref TCM_Recall (int refID)
{
  if (refID < 0) {
    return NullRef;
  } else {
    return GET_TCM_GLOBAL(rootNode)->recall(refID);
  }
}

/**************************************************************************
 *
 * FUNCTION: void TCM_Forget (int refID)
 *
 * DESCRIPTION: Un-memoize the node, freeing up storage
 *
 **************************************************************************/
void TCM_Forget (int refID)
{
  if (refID >= 0) {
    GET_TCM_GLOBAL(rootNode)->forget(refID);
  }
}

/**************************************************************************
 *
 * FUNCTION: void TCM_Forget (int refID)
 *
 * DESCRIPTION: Combines TCM_Recall & TCM_Forget
 *
 **************************************************************************/

TCM_Task_Tree_Ref TCM_Retrieve (int refID)
{
  if (refID < 0) {
    return NullRef;
  } else {
    return GET_TCM_GLOBAL(rootNode)->retrieve(refID);
  }
}

/*****************************************************************
 * 
 * Functions having to do with temporal constraints
 *
 *****************************************************************/

BOOLEAN TCM_Point::valid (void) const
{ return validRef(_interval._ref); }

TCM_Interval TCM_HandlingOf (const TCM_Task_Tree_Ref &nodeRef)
{
  TCM_Interval interval(Handling_Interval, nodeRef);
  return interval;
}

TCM_Interval TCM_PlanningOf (const TCM_Task_Tree_Ref &nodeRef)
{
  TCM_Interval interval(Planning_Interval, nodeRef);
  return interval;
}

TCM_Interval TCM_AchievingOf (const TCM_Task_Tree_Ref &nodeRef)
{
  TCM_Interval interval(Achieving_Interval, nodeRef);
  return interval;
}

TCM_Point TCM_StartOf (const TCM_Interval &interval)
{
  TCM_Point point(Start_Point, interval);
  return point;
}

TCM_Point TCM_EndOf (const TCM_Interval &interval)
{
  TCM_Point point(End_Point, interval);
  return point;
}

TCM_Return_Type TCM_AddConstraint (const TCM_Task_Tree_Ref &nodeRef,
				   int tplConstraints)
{
  RETURN_IF_INVALID_NODE(nodeRef, "TCM_AddConstraint");

  if (checkInconsistent(tplConstraints)) {
    tcmError("TCM_AddConstraint: Inconsistent temporal constraints %d\n",
	     tplConstraints);
    return TCM_Error;
  } else {
    nodeRef->addConstraints(tplConstraints);
    return TCM_Ok;
  }
}

/**************************************************************************
 *
 * FUNCTION: TCM_Return_Type TCM_Serialize (const TCM_Task_Tree_Ref &beforeRef,
 *			                    const TCM_Task_Tree_Ref &afterRef)
 *
 * DESCRIPTION: Constrain the end of planning of "beforeRef" to be before 
 *              the start of planning of "afterRef", and the same for their
 *              achievement intervals.
 *
 **************************************************************************/
TCM_Return_Type TCM_Serialize (const TCM_Task_Tree_Ref &beforeRef,
			       const TCM_Task_Tree_Ref &afterRef)
{
  RETURN_IF_INVALID_NODE(beforeRef, "TCM_Serialize");
  RETURN_IF_INVALID_NODE(afterRef, "TCM_Serialize");

  Time_Point end_ach_before(End_Point, Achieving_Interval, beforeRef);
  Time_Point start_plan_after(Start_Point, Planning_Interval, afterRef);
  Time_Point start_ach_after(Start_Point, Achieving_Interval, afterRef);

  constrainBefore(end_ach_before, start_plan_after);
  constrainBefore(end_ach_before, start_ach_after);

  return TCM_Ok;
}

TCM_Return_Type TCM_DelayUntil (TCM_Point const &timePoint,
				TCM_Point const &delayTimePoint)
{
  RETURN_IF_INVALID_TIME(timePoint, "TCM_DelayUntil");
  RETURN_IF_INVALID_TIME(delayTimePoint, "TCM_DelayUntil");

  Time_Point t1(timePoint), t2(delayTimePoint);
  constrainBefore(t2, t1);
  return TCM_Ok;
}

TCM_Return_Type TCM_DelayUntil (TCM_Point const &timePoint,
				MSecs absoluteTime)
{
  RETURN_IF_INVALID_TIME(timePoint, "TCM_DelayUntil");

  Time_Point point(timePoint);
  constrainWhen(point, absoluteTime);

  return TCM_Ok;
}

TCM_Return_Type TCM_DelayFor (TCM_Point const &timePoint, MSecs relativeWait)
{
  RETURN_IF_INVALID_TIME(timePoint, "TCM_DelayFor");

  return TCM_DelayUntil(timePoint, absoluteTime(relativeWait)); 
}

/**************************************************************************
 *
 * FUNCTION: TCM_Return_Type TCM_DelayForAfter (TCM_Point const &timePoint,
 *				                MSecs relativeWait,
 *                                              TCM_Point const &delayTimePoint)
 *
 * DESCRIPTION: Delay this timepoint for "relativeWait" msecs after 
 *              the "delayTimePoint" has passed.
 *
 **************************************************************************/

TCM_Return_Type TCM_DelayForAfter (TCM_Point const &timePoint,
				   MSecs relativeWait,
				   TCM_Point const &delayTimePoint)
{
  RETURN_IF_INVALID_TIME(timePoint, "TCM_DelayForAfter");
  RETURN_IF_INVALID_TIME(delayTimePoint, "TCM_DelayForAfter");

  Time_Point beforePoint(timePoint), afterPoint(delayTimePoint);

  constrainWhenAfter(beforePoint.nodeOf(), beforePoint.signalOf(),
		     afterPoint, relativeWait);

  return TCM_Ok;
}

/**************************************************************************
 *
 * FUNCTION: TCM_Return_Type TCM_DelayUntilSignal (TCM_Point const &timePoint)
 *
 * DESCRIPTION: Don't enable this timepoint until TCM_Signal is invoked
 *              for that timepoint.
 *
 **************************************************************************/
TCM_Return_Type TCM_DelayUntilSignal (TCM_Point const &timePoint)
{
  RETURN_IF_INVALID_TIME(timePoint, "TCM_DelayUntilSignal");
  
  Time_Point time(timePoint);
  Task_Tree_Ref node = time.nodeOf();
  
  node->addExpectedEvent(time.signalOf(), node);
  return TCM_Ok;
}

/**************************************************************************
 *
 * FUNCTION: TCM_Return_Type TCM_Signal (TCM_Point const &timePoint)
 *
 * DESCRIPTION: Send an explicit signal to enable this timepoint.
 *
 **************************************************************************/
TCM_Return_Type TCM_Signal (TCM_Point const &timePoint)
{
  RETURN_IF_INVALID_TIME(timePoint, "TCM_Signal");
  
  Time_Point time(timePoint);
  Task_Tree_Ref node = time.nodeOf();
  Event event(time.signalOf(), node);
  
  node->signal(event);
  return TCM_Ok;
}

TCM_Return_Type TCM_WaitUntil (TCM_Point const &waitTimePoint)
{
  RETURN_IF_INVALID_TIME(waitTimePoint, "TCM_WaitUntil");

  if (!TCM_IsPast(waitTimePoint)) {
    Time_Point timepoint(waitTimePoint);
    GET_TCM_GLOBAL(agenda).clearQueues(&timepoint, INFINITE_TIME, TRUE);
  }
  return TCM_Ok;
}

TCM_Return_Type TCM_WaitUntil (MSecs absoluteTime)
{
  if (!TCM_IsPast(absoluteTime)) {
    GET_TCM_GLOBAL(agenda).clearQueues(NULL, absoluteTime, TRUE);
  }
  return TCM_Ok;
}

TCM_Return_Type TCM_WaitFor (MSecs relativeWait)
{
  return TCM_WaitUntil(absoluteTime(relativeWait));
}

BOOLEAN TCM_IsPast (TCM_Point const &timePoint)
{
  Time_Point tp(timePoint);
  return tp.isPast();
}

BOOLEAN TCM_IsPast (MSecs absoluteTime)
{
  return (absoluteTime <= timeInMsecs());
}

// The way this works is that the command node does not really finish
// until the timer sends it a "Done_Handling_Signal"
static void delayHnd (const TCM_Task_Tree_Ref &node, const void *data)
{
  tcmMessage("  Invoking DelayHnd\n");

  MSecs delay = (MSecs)data;
  addTimedEvent(delay, node, Done_Handling_Signal);

  TCM_Success(node);
}

// Assumes that "Time" and "ptr" are same length.
// Safer would be to create (and destroy) the data, and pass a true pointer.
TCM_Task_Tree_Ref TCM_CreateDelayCommand (const TCM_Task_Tree_Ref &parentRef,
					  MSecs relativeWait)
{
  RETURN_ON_INVALID_NODE(parentRef, "TCM_CreateDelayCommand", NullRef);

  MSecs time = absoluteTime(relativeWait);

  return TCM_CreateCommandNode(parentRef, "tcmDelay",
			       new Callback_Action(delayHnd, (void *)time));
}


/*****************************************************************
 * 
 * Functions having to do with traversing task trees.
 *
 *****************************************************************/

TCM_Task_Tree_Ref TCM_Parent (const TCM_Task_Tree_Ref &childRef)
{
  RETURN_ON_INVALID_NODE(childRef, "TCM_Parent", NullRef);

  return childRef->getParent();
}

TCM_Task_Tree_Ref TCM_FirstChild (const TCM_Task_Tree_Ref &parentRef)
{
  RETURN_ON_INVALID_NODE(parentRef, "TCM_FirstChild", NullRef);

  return parentRef->firstChild();
}

TCM_Task_Tree_Ref TCM_LastChild (const TCM_Task_Tree_Ref &parentRef)
{
  RETURN_ON_INVALID_NODE(parentRef, "TCM_LastChild", NullRef);

  return parentRef->lastChild(FALSE);
}

TCM_Task_Tree_Ref TCM_NextChild (const TCM_Task_Tree_Ref &childRef)
{
  RETURN_ON_INVALID_NODE(childRef, "TCM_NextChild", NullRef);

  return childRef->getParent()->nextChild(childRef);
}

TCM_Task_Tree_Ref TCM_PreviousChild (const TCM_Task_Tree_Ref &childRef)
{
  RETURN_ON_INVALID_NODE(childRef, "TCM_PreviousChild", NullRef);

  return childRef->getParent()->previousChild(childRef);
}

TCM_Task_Tree_Ref TCM_ChildNamed (const TCM_Task_Tree_Ref &parentRef,
				  STRING childName)
{
  RETURN_ON_INVALID_NODE(parentRef, "TCM_ChildNamed", NullRef);
  
  Task_Tree_Node_Ptr child = parentRef->childNamed(childName);
  
  return (child ? makeHandle(child) : NullRef);
}

TCM_Task_Tree_Ref TCM_AncestorNamed (const TCM_Task_Tree_Ref &descendantRef,
				     STRING ancestorName)
{
  RETURN_ON_INVALID_NODE(descendantRef, "TCM_AncestorNamed", NullRef);
  
  Task_Tree_Node_Ptr ancestor = descendantRef->ancestorNamed(ancestorName);
  
  return (ancestor ? makeHandle(ancestor) : NullRef);
}

STRING TCM_NodeName (const TCM_Task_Tree_Ref &nodeRef)
{
  RETURN_ON_INVALID_NODE(nodeRef, "TCM_NodeName", NULL);
  return nodeRef->getName();
}

/*****************************************************************
 * 
 * Functions having to do with terminating and suspending nodes
 *
 *****************************************************************/

TCM_Return_Type TCM_TerminateNode (const TCM_Task_Tree_Ref &nodeRef)
{
  RETURN_IF_INVALID_NODE(nodeRef, "TCM_TerminateNode");

  return (nodeRef->terminate() ? TCM_Ok : TCM_Error);
}

TCM_Return_Type TCM_SuspendNode (const TCM_Task_Tree_Ref &nodeRef)
{
  RETURN_IF_INVALID_NODE(nodeRef, "TCM_SuspendNode");

  nodeRef->suspend();
  return TCM_Ok;
}

TCM_Return_Type TCM_UnsuspendNode (const TCM_Task_Tree_Ref &nodeRef)
{
  RETURN_IF_INVALID_NODE(nodeRef, "TCM_UnsuspendNode");

  nodeRef->unsuspend();
  return TCM_Ok;
}

BOOLEAN TCM_IsSuspended (const TCM_Task_Tree_Ref &nodeRef)
{
  return (checkNodeValidity(nodeRef, "TCM_IsSuspended") &&
	  nodeRef->isSuspended());
}

TCM_Return_Type TCM_TerminateAt (const TCM_Task_Tree_Ref &nodeRef,
				 TCM_Point const &terminateTimePoint)
{
  RETURN_IF_INVALID_NODE(nodeRef, "TCM_TerminateAt");
  RETURN_IF_INVALID_TIME(terminateTimePoint, "TCM_TerminateAt");

  Time_Point point(terminateTimePoint);
  terminateAfter(nodeRef, point);

  return TCM_Ok;
}

TCM_Return_Type TCM_TerminateAt (const TCM_Task_Tree_Ref &nodeRef,
				 MSecs absoluteTime)
{
  RETURN_IF_INVALID_NODE(nodeRef, "TCM_TerminateAt");

  terminateWhen(nodeRef, absoluteTime);
  return TCM_Ok;
}

TCM_Return_Type TCM_TerminateIn (const TCM_Task_Tree_Ref &nodeRef,
				 MSecs relativeWait)
{
  RETURN_IF_INVALID_NODE(nodeRef, "TCM_TerminateIn");

  return TCM_TerminateAt(nodeRef, absoluteTime(relativeWait));
}

// Terminate this node in "relativeWait" msecs after 
//  the "terminateTimePoint" have passed.
TCM_Return_Type TCM_TerminateInAfter (const TCM_Task_Tree_Ref &nodeRef,
				      MSecs relativeWait,
				      TCM_Point const &terminateTimePoint)
{
  RETURN_IF_INVALID_NODE(nodeRef, "TCM_TerminateInAfter");
  RETURN_IF_INVALID_TIME(terminateTimePoint, "TCM_TerminateInAfter");

  Time_Point point(terminateTimePoint);

  constrainWhenAfter(nodeRef, Start_Terminating_Signal, point, relativeWait);

  return TCM_Ok;
}

TCM_Return_Type TCM_OnTermination (const TCM_Task_Tree_Ref &nodeRef,
				   const TCM_Task_Tree_Ref &terminationRef)
{
  RETURN_IF_INVALID_NODE(nodeRef, "TCM_OnTermination");
  RETURN_IF_INVALID_NODE(terminationRef, "TCM_OnTermination");

  nodeRef->addTerminationAction(terminationRef);

  return TCM_Ok;
}

TCM_Return_Type TCM_OnTermination (const TCM_Task_Tree_Ref &nodeRef,
				   STRING name,
				   const TCM_Action_Ref &terminationAction)
{
  return TCM_OnTermination(nodeRef, 
			   new Task_Tree_Goal_Node(name, terminationAction));
}

TCM_Return_Type TCM_OnTermination (const TCM_Task_Tree_Ref &nodeRef,
				   STRING name,
				   const void *callbackData,
				   CALLBACK_FN_TYPE callback)
{
  return TCM_OnTermination(nodeRef, name, 
			   new Callback_Action(callback, callbackData));
}

/*****************************************************************
 * 
 * Functions having to do with exceptions
 *
 *****************************************************************/

TCM_Return_Type TCM_AddExceptionHandler(const TCM_Task_Tree_Ref &nodeRef,
					STRING type,
					const TCM_Action_Ref &action,
					unsigned int maxInvocations)
{
  RETURN_IF_INVALID_NODE(nodeRef, "TCM_AddExceptionHandler");

  addExceptionHandler(nodeRef, type, action, maxInvocations);

  return TCM_Ok;
}

TCM_Return_Type TCM_AddExceptionHandler(const TCM_Task_Tree_Ref &nodeRef,
					STRING type,
					const CALLBACK_FN_TYPE callback,
					unsigned int maxInvocations)
{
  return TCM_AddExceptionHandler(nodeRef, type, new Callback_Action(callback),
				 maxInvocations);
}

TCM_Return_Type TCM_RemoveExceptionHandler(const TCM_Task_Tree_Ref &nodeRef,
					   STRING type,
					   const TCM_Action_Ref &action)
{
  RETURN_IF_INVALID_NODE(nodeRef, "TCM_RemoveExceptionHandler");

  removeExceptionHandler(nodeRef, type, action);

  return TCM_Ok;
}

TCM_Return_Type TCM_RemoveExceptionHandler(const TCM_Task_Tree_Ref &nodeRef,
					   STRING type,
					   const CALLBACK_FN_TYPE callback)
{
  Callback_Action action(callback);

  return TCM_RemoveExceptionHandler(nodeRef, type, &action);
}

TCM_Return_Type TCM_Bypass(const TCM_Task_Tree_Ref &nodeRef)
{
  RETURN_IF_INVALID_NODE(nodeRef, "TCM_Bypass");

  if (nodeRef->isException()) {
    ((Exception_Node *)(*nodeRef))->bypass();
    return TCM_Ok;
  } else {
    tcmError("TCM_bypass: %s is not an exception node\n", nodeRef->getName());
    return TCM_Error;
  }
}

TCM_Return_Type TCM_Retry(const TCM_Task_Tree_Ref &nodeRef,
			  const void *retryData)
{
  UNUSED(retryData);

  RETURN_IF_INVALID_NODE(nodeRef, "TCM_Retry");  

  tcmWarning("TCM_Retry: Not Yet Implemented\n");
  return TCM_Ok;
}

const void *TCM_FailureData (const TCM_Task_Tree_Ref &nodeRef)
{
  RETURN_ON_INVALID_NODE(nodeRef, "TCM_FailureData", NULL);
	
  void const *failureData = nodeRef->getNodeData();

  return (failureData != NULL ? failureData : nodeRef->getNodeRefCountData());
}

/*****************************************************************
 * 
 * Functions having to do with monitors
 *
 *****************************************************************/

static STRING monitorActivationName (Task_Tree_Monitor_Node *monitor,
				     STRING activationName)
{
  static char actName[100];

  if (activationName) {
    return activationName;
  } else {
    sprintf(actName, "ACT-%s", monitor->getName());
    return actName;
  }
}


/**************************************************************************
 *
 * FUNCTION: TCM_Task_Tree_Ref TCM_AllocateMonitorNode (STRING name)
 *
 * DESCRIPTION: Create a new monitor node, with the given name.
 *
 * NOTE: The node is not connected to the task tree, and is not runnable.
 *
 **************************************************************************/

TCM_Task_Tree_Ref TCM_AllocateMonitorNode (STRING name)
{
  return new Task_Tree_Monitor_Node(name);
}

/**************************************************************************
 *
 * FUNCTION: TCM_Task_Tree_Ref TCM_SetMonitorParameters (...)
 *
 * DESCRIPTION: Set all the parameters for the (regular) monitor.
 *
 * NOTE: The node is still not connected to the task tree, and is not runnable.
 *
 **************************************************************************/

TCM_Return_Type 
TCM_SetMonitorParameters (const TCM_Task_Tree_Ref &monitorRef,
			  const TCM_Action_Ref &activationAction,
			  unsigned int maxActivations,
			  unsigned int maxTriggers,
			  STRING activationName,
			  int activationConstraints)
{
  if (!monitorRef->isMonitor()) {
    tcmWarning("TCM_SetMonitorParameters: %s is not a monitor\n",
	       monitorRef->getName());
    return TCM_Error;
  } else {
    Task_Tree_Monitor_Node *monitor = ((Task_Tree_Monitor_Node *)*monitorRef);
    Monitor_Action *action
      = new Monitor_Action(activationAction,
			   monitorActivationName(monitor, activationName),
			   maxActivations, activationConstraints);

    monitor->setAction(action);
    monitor->setMaxTriggers(maxTriggers);

    return TCM_Ok;
  }
}

/**************************************************************************
 *
 * FUNCTION: TCM_Task_Tree_Ref TCM_AllocatePollingMonitorNode (STRING name)
 *
 * DESCRIPTION: Create a new polling monitor node, with the given name.
 *
 * NOTE: The node is not connected to the task tree, and is not runnable.
 *
 **************************************************************************/

TCM_Task_Tree_Ref TCM_AllocatePollingMonitorNode (STRING name)
{
  return new Task_Tree_Monitor_Node(name);
}

/**************************************************************************
 *
 * FUNCTION: TCM_Task_Tree_Ref TCM_SetPollingMonitorParameters (...)
 *
 * DESCRIPTION: Set all the parameters for the (regular) monitor.
 *
 * NOTE: The node is still not connected to the task tree, and is not runnable.
 *
 **************************************************************************/

TCM_Return_Type 
TCM_SetPollingMonitorParameters (const TCM_Task_Tree_Ref &monitorRef,
				 const TCM_Action_Ref &activationAction,
				 MSecs period,
				 unsigned int maxActivations,
				 unsigned int maxTriggers,
				 STRING activationName,
				 int activationConstraints,
				 BOOLEAN initialWait)
{
  if (!monitorRef->isMonitor()) {
    tcmWarning("TCM_SetPollingMonitorParameters: %s is not a monitor\n",
	       monitorRef->getName());
    return TCM_Error;
  } else {
    Task_Tree_Monitor_Node *monitor = ((Task_Tree_Monitor_Node *)*monitorRef);
    Polling_Monitor_Action *action 
      = new Polling_Monitor_Action(activationAction,
				   monitorActivationName(monitor, 
							 activationName),
				   period, maxActivations,
				   activationConstraints, initialWait);

    monitor->setAction(action);
    monitor->setMaxTriggers(maxTriggers);

    return TCM_Ok;
  }
}

TCM_Task_Tree_Ref TCM_CreateMonitorNode (const TCM_Task_Tree_Ref &parentRef,
					 STRING name, 
					 const TCM_Action_Ref &activationAction,
					 unsigned int maxActivations,
					 unsigned int maxTriggers,
					 STRING activationName,
					 int activationConstraints)
{
  RETURN_ON_INVALID_NODE(parentRef, "TCM_CreateMonitorNode", NullRef);

  TCM_Task_Tree_Ref monitorRef = TCM_AllocateMonitorNode(name);

  TCM_SetMonitorParameters(monitorRef, activationAction, maxActivations,
			   maxTriggers, activationName, activationConstraints);

  taskTreeInsert(*parentRef, *monitorRef, TRUE);

  return monitorRef;
}

TCM_Task_Tree_Ref TCM_CreateMonitorNode (const TCM_Task_Tree_Ref &parentRef,
					 STRING name, 
					 const void *callbackData,
					 CALLBACK_FN_TYPE activationCallback,
					 unsigned int maxActivations,
					 unsigned int maxTriggers,
					 STRING activationName,
					 int activationConstraints)
{
  return TCM_CreateMonitorNode(parentRef, name,
			       new Callback_Action(activationCallback,
						   callbackData),
			       maxActivations, maxTriggers,
			       activationName, activationConstraints);
}

TCM_Task_Tree_Ref
TCM_CreatePollingMonitorNode (const TCM_Task_Tree_Ref &parentRef,
			      STRING name, 
			      const TCM_Action_Ref &activationAction,
			      MSecs period,
			      unsigned int maxActivations,
			      unsigned int maxTriggers,
			      STRING activationName,
			      int activationConstraints,
			      BOOLEAN initialWait)
{
  RETURN_ON_INVALID_NODE(parentRef, "TCM_CreatePollingMonitorNode", NullRef);

  TCM_Task_Tree_Ref monitorRef = TCM_AllocateMonitorNode(name);

  TCM_SetPollingMonitorParameters(monitorRef, activationAction, period,
				  maxActivations, maxTriggers, activationName,
				  activationConstraints, initialWait);

  taskTreeInsert(*parentRef, *monitorRef, TRUE);

  return monitorRef;
}

TCM_Task_Tree_Ref
TCM_CreatePollingMonitorNode (const TCM_Task_Tree_Ref &parentRef,
			      STRING name, 
			      const void *callbackData,
			      CALLBACK_FN_TYPE activationCallback,
			      MSecs period,
			      unsigned int maxActivations,
			      unsigned int maxTriggers,
			      STRING activationName,
			      int activationConstraints,
			      BOOLEAN initialWait)
{
  return TCM_CreatePollingMonitorNode(parentRef, name,
				      new Callback_Action(activationCallback,
							  callbackData),
				      period, maxActivations, maxTriggers,
				      activationName, activationConstraints,
				      initialWait);
}

TCM_Return_Type TCM_Trigger (const TCM_Task_Tree_Ref &monitorRef)
{
  RETURN_IF_INVALID_NODE(monitorRef, "TCM_Trigger");
  Task_Tree_Monitor_Node *monitor = monitorDereference(monitorRef);

  if (monitor) {
    monitor->trigger();
    return TCM_Ok;
  } else {
    tcmWarning("TCM_Trigger: Node is not a monitor\n");
    return TCM_Error;
  }
}

TCM_Return_Type TCM_Activate (const TCM_Task_Tree_Ref &monitorRef,
			      const void *activationData)
{
  RETURN_IF_INVALID_NODE(monitorRef, "TCM_Activate");
  Task_Tree_Monitor_Node *monitor = monitorDereference(monitorRef);

  if (monitor) {
    monitor->activate(activationData);
    return TCM_Ok;
  } else {
    tcmWarning("TCM_Activate: Node is not a monitor\n");
    return TCM_Error;
  }
}

TCM_Return_Type TCM_ActivateAt (const TCM_Task_Tree_Ref &monitorRef,
				TCM_Point const &activateTimePoint)
{
  RETURN_IF_INVALID_NODE(monitorRef, "TCM_ActivateAt");
  Task_Tree_Node *monitor = monitorDereference(monitorRef);

  if (monitor) {
    Time_Point point(activateTimePoint);
    addEventPair(point.nodeOf(), point.stateOf(), Activate_Signal, monitorRef);
    return TCM_Ok;
  } else {
    tcmWarning("TCM_ActivateAt: Node is not a monitor\n");
    return TCM_Error;
  }
}

TCM_Return_Type TCM_ActivateAt (const TCM_Task_Tree_Ref &monitorRef,
				 MSecs absoluteTime)
{
  RETURN_IF_INVALID_NODE(monitorRef, "TCM_ActivateAt");
  Task_Tree_Node *monitor = monitorDereference(monitorRef);

  if (monitor) {
    addTimedEvent(absoluteTime, monitorRef, Activate_Signal);
    return TCM_Ok;
  } else {
    tcmWarning("TCM_ActivateAt: Node is not a monitor\n");
    return TCM_Error;
  }
}

TCM_Return_Type TCM_ActivateIn (const TCM_Task_Tree_Ref &monitorRef,
				 MSecs relativeWait)
{
  RETURN_IF_INVALID_NODE(monitorRef, "TCM_ActivateIn");

  return TCM_ActivateAt(monitorRef, absoluteTime(relativeWait));
}

// Activate this node in "relativeWait" msecs after 
//  the "activateTimePoint" have passed.
TCM_Return_Type TCM_ActivateInAfter (const TCM_Task_Tree_Ref &monitorRef,
				     MSecs relativeWait,
				     TCM_Point const &activateTimePoint)
{
  RETURN_IF_INVALID_NODE(monitorRef, "TCM_ActivateInAfter");
  RETURN_IF_INVALID_TIME(activateTimePoint, "TCM_ActivateInAfter");

  Time_Point point(activateTimePoint);

  constrainWhenAfter(monitorRef, Activate_Signal, point, relativeWait);

  return TCM_Ok;
}

const void *TCM_ActivationData (const TCM_Task_Tree_Ref &monitorRef)
{
  RETURN_ON_INVALID_NODE(monitorRef, "TCM_ActivationData", NULL);
  Task_Tree_Node *monitor = monitorDereference(monitorRef);

  if (monitor) {
    return monitorRef->getNodeData();
  } else {
    tcmWarning("TCM_ActivationData: Node is not a monitor\n");
    return NULL;
  }
}

unsigned int TCM_NumTriggers (const TCM_Task_Tree_Ref &nodeRef)
{
  RETURN_ON_INVALID_NODE(nodeRef, "TCM_NumTriggers", 0);
  Task_Tree_Monitor_Node *monitor = monitorDereference(nodeRef);

  if (monitor) {
    return monitor->numTriggers();
  } else {
    tcmWarning("TCM_NumTriggers: Node is not a monitor\n");
    return 0;
  }
}

unsigned int TCM_NumActivations (const TCM_Task_Tree_Ref &nodeRef)
{
  RETURN_ON_INVALID_NODE(nodeRef, "TCM_NumActivations", 0);
  Task_Tree_Monitor_Node *monitor = monitorDereference(nodeRef);

  if (monitor) {
    return monitor->numActivations();
  } else {
    tcmWarning("TCM_NumActivations: Node is not a monitor\n");
    return 0;
  }  
}


/*****************************************************************
 * 
 * Functions to invoke functions (not adding tree nodes) in 
 * response to particular events.
 *
 *****************************************************************/

TCM_Return_Type TCM_InvokeWhen (TCM_Point const &timePoint,
				SIMPLE_CALLBACK_FN_TYPE callback,
				const void *callbackData)
{
  RETURN_IF_INVALID_TIME(timePoint, "TCM_InvokeWhen");
  RETURN_IF_NULL(callback, "TCM_InvokeWhen");
  
  Time_Point point(timePoint);
  point.nodeOf()->addRequestedEvent(point.stateOf(), callback, callbackData);
  return TCM_Ok;
}

TCM_Return_Type TCM_InvokeWhen (MSecs absoluteTime,
				SIMPLE_CALLBACK_FN_TYPE callback,
				const void *callbackData)
{
  RETURN_IF_NULL(callback, "TCM_InvokeWhen");
  
  addTimedEvent(absoluteTime, callback, callbackData);
  return TCM_Ok;
}

TCM_Return_Type TCM_InvokeAfter (MSecs relativeWait,
				 SIMPLE_CALLBACK_FN_TYPE callback,
				 const void *callbackData)
{
  RETURN_IF_NULL(callback, "TCM_InvokeAfter");

  return TCM_InvokeWhen(absoluteTime(relativeWait), callback, callbackData);
}

/*****************************************************************
 * 
 * Functions for interacting with TCM
 *
 *****************************************************************/

TCM_Return_Type TCM_Initialize (void)
{
  initializeGlobalState();

  tcmStartTerminalLogging();

  return TCM_Ok;
}

/**************************************************************************
 *
 * FUNCTION: BOOLEAN isCurrentState(State_Enum state)
 *
 * DESCRIPTION: If "waitingAllowed" is true, TCM_ProcessAgenda can block
 *              on timeout signals and/or external events.  Otherwise,
 *              returns when all pending internal events have been processed.
 *
 **************************************************************************/

TCM_Return_Type TCM_ProcessAgenda (BOOLEAN waitingAllowed,
				   MSecs   relativeTimeout /*=INFINITE_TIME*/ )
{
  GET_TCM_GLOBAL(agenda).clearQueues(NULL, absoluteTime( relativeTimeout ),
				     waitingAllowed );
  return TCM_Ok;
}

TCM_Return_Type TCM_AddExternalEvent (int sd,
				      SD_CALLBACK_FN_TYPE callback,
				      const void *callbackData)
{
  if (sd >= 0) {
    GET_TCM_GLOBAL(agenda).addExternalEvent(sd, callback, callbackData);
    return TCM_Ok;
  } else {
    tcmError("TCM_AddExternalEvent: Illegal file descriptor %d\n", sd);
    return TCM_Error;
  }
}
	
TCM_Return_Type TCM_RemoveExternalEvent (int sd)
{
  if (sd >= 0) {
    GET_TCM_GLOBAL(agenda).removeExternalEvent(sd);
    return TCM_Ok;
  } else {
    tcmError("TCM_RemoveExternalEvent: Illegal file descriptor %d\n", sd);
    return TCM_Error;
  }
}

void TCM_SetExternalEventBypassHandler (
			 EXTERNAL_EVENT_BYPASS_HANDLER_TYPE theBypassHandler )
{
  GET_TCM_GLOBAL(agenda).setExternalEventBypassHandler ( theBypassHandler );
}


TCM_Return_Type TCM_DisplayTree(FILE *stream, const TCM_Task_Tree_Ref &nodeRef)
{
  RETURN_IF_INVALID_NODE(nodeRef, "TCM_DisplayTree");

  nodeRef->displayTree(stream, 0);

  return TCM_Ok;
}

static void setLoggingOptions(Logger &logger, int options)
{
  logger.setMessageLogging(options & Log_Msg);
  logger.setStatusLogging(options & Log_Status);
  logger.setTimeLogging(options & Log_Time);
  logger.setIdLogging(options & Log_Id);
  logger.setParentIdLogging(options & Log_ParentId);
  logger.setIgnoreLogging(options & Log_Ignore);
}

static int getLoggingOptions(Logger &logger)
{
  int options = Log_None;

  if (logger.message()) options |= Log_Msg;
  if (logger.status()) options |= Log_Status;
  if (logger.time()) options |= Log_Time;
  if (logger.id()) options |= Log_Id;
  if (logger.parentId()) options |= Log_ParentId;
  if (logger.ignore()) options |= Log_Ignore;

  return options;
}

void TCM_SetTerminalLoggingOptions (int options)
{
  setLoggingOptions(GET_TCM_GLOBAL(terminalLogger), options);
}

void TCM_SetFileLoggingOptions (int options)
{
  setLoggingOptions(GET_TCM_GLOBAL(fileLogger), options);
}

int TCM_TerminalLoggingOptions (void)
{
  return getLoggingOptions(GET_TCM_GLOBAL(terminalLogger));
}

int TCM_FileLoggingOptions (void)
{
  return getLoggingOptions(GET_TCM_GLOBAL(fileLogger));
}
