/**************************************************************************
 * 
 * PROJECT: Task Control Management.
 *
 * (c) Copyright 1999 Reid Simmons.  All rights reserved.
 *
 * FILE: virtual.cc
 *
 * ABSTRACT: Defines "virtual" task tree nodes used for distributing a
 *           single task tree amongst multiple processes.  A "virtual" node
 *           has "real" counterpart in another process.  "Virtual" nodes 
 *           know how to send messages to their real counterparts and how
 *           to interpret messages sent to them.
 *
 *   Example:      Process 1      Process 2
 *                   [A]             [A']
 *                   / \               \
 *                 [B] [C']            [C]
 *
 *       In process 1, nodes A and B are real, and C' is virtual;
 *       In process 2, node C is real and node A' is virtual.  Virtual
 *       Virtual node C' corresponds to real node C, and A' corresponds to A.
 *
 * $Revision: 1.1 $
 * $Date: 1999/08/04 14:00:20 $
 * $Author: reids $
 *
 * REVISION HISTORY:
 *
 * $Log: virtual.cc,v $
 * Revision 1.1  1999/08/04 14:00:20  reids
 * Distributed version of TCM.  Provides the ability for one process to
 *  add tasks, constraints, exception handlers, etc. with another process.
 *  The distributed version is created by compiling with -DDISTRIBUTED flag,
 *   which can be gotten by "gmaking" with DISTRIBUTED=1.
 *
 **************************************************************************/

#include "virtual.h"
#include "distributed.h"
#include "agenda.h"

Task_Tree_Node_List Task_Tree_Node::virtualNodes;

void Virtual_Node::_initialize (void)
{
  // Undo the effects of the Task_Tree_Node initialization
  Task_Tree_Node::removeExpectedEvent(Done_Handling_Signal, this);
}

static void setVirtualAddressIfNeeded (Task_Tree_Ref const &node)
{
  if (!node->validVirtualAddress()) {
    node->setVirtualAddress(thisAgent(), thisAgent(), node->getNodeId());
  }
}

void Virtual_Node::addParent (Task_Tree_Ref const &parentNode,
			      BOOLEAN canBeLastChild, BOOLEAN addConstraints)
{
  setVirtualAddressIfNeeded(parentNode);
  distributedAddParent(this, parentNode, canBeLastChild);
  Task_Tree_Node::addParent(parentNode, canBeLastChild, FALSE);
}

void Virtual_Node::addChild (Task_Tree_Ref const &childNode, BOOLEAN addConstraints)
{
  setVirtualAddressIfNeeded(childNode);
  distributedAddChild(this, childNode);
  Task_Tree_Node::addChild(childNode, FALSE);
}

void Virtual_Node::removeChild (Task_Tree_Ref const &childNode)
{
  setVirtualAddressIfNeeded(childNode);
  distributedRemoveChild(this, childNode);
}

void Virtual_Node::signal (const Event &event)
{
  setVirtualAddressIfNeeded(event.getNode());
  distributedSignal(this, event);
}

void Virtual_Node::addExpectedEvent (Signal_Enum signal,
				     Task_Tree_Ref const &signallingNode,
				     const void *timerData)
{
  setVirtualAddressIfNeeded(signallingNode);
  distributedAddExpectedEvent(this, signal, signallingNode, timerData);
}

void Virtual_Node::addRequestedEvent (State_Enum state, Signal_Enum signal,
				      Task_Tree_Ref const &signalledNode)
{
  setVirtualAddressIfNeeded(signalledNode);
  distributedAddRequestedEvent(this, state, signal, signalledNode);
}

// The idea here is to add the signal to the distributed node, but 
//   add the callback and data here.
void Virtual_Node::addRequestedEvent(State_Enum state, 
				     SIMPLE_CALLBACK_FN_TYPE callbackFn, 
				     const void *data)
{
  State_Event event(callbackFn, data, state);

  requestedEvents.insert(event);
  distributedAddRequestedCallback(this, state);
}

void Virtual_Node::invokeCallback (State_Enum state)
{
  State_Event_Iterator listIter(&requestedEvents);

  for (listIter.reset(); listIter; listIter++) {
    if (state == listIter()->getState()) {
#ifdef TRACE_VIRTUAL
      tcmWarning("invokeCallback for %s\n", stateName(state));
#endif
      listIter()->sendSignal(this);
      listIter.removeCurrent();
    }
  }
}

void Virtual_Node::removeExpectedEvent(Signal_Enum signal,
				       Task_Tree_Ref const &signallingNode)
{
  setVirtualAddressIfNeeded(signallingNode);
  distributedRemoveExpectedEvent(this, signal, signallingNode);
}

void Virtual_Node::removeRequestedEvents (Signal_Enum signal,
					  Task_Tree_Ref const &signalledNode)
{
  setVirtualAddressIfNeeded(signalledNode);
  distributedRemoveRequestedEvents(this, signal, signalledNode);
}

BOOLEAN Virtual_Node::terminate (void)
{
  distributedTerminate(this);
  // The "regular" task tree nodes check for various error conditions, but
  // that is too difficult, and hardly worthwhile doing, for distributed nodes.
  return TRUE;
}

BOOLEAN Virtual_Node::deallocate (void)
{
  distributedDeallocate(this);
  // The "regular" task tree nodes check for various error conditions, but
  // that is too difficult, and hardly worthwhile doing, for distributed nodes.
  return TRUE;
}

void Virtual_Node::suspend (void)
{
  distributedSuspend(this);
}

void Virtual_Node::unsuspend (void)
{
  distributedUnsuspend(this);
}

void Virtual_Node::delayTermination (void)
{
  distributedDelayTermination(this);
}

void Virtual_Node::enableTermination (void)
{
  distributedEnableTermination(this);
}

void 
Virtual_Node::addTerminationAction (const TCM_Task_Tree_Ref &terminationRef)
{
  setVirtualAddressIfNeeded(terminationRef);
  distributedAddTermination(this, terminationRef);
}

void Virtual_Node::displayTree (FILE *stream, int position = 0) const
{
  // You can only display the distributed tree to either stdout or stderr
  distributedDisplayTree((Task_Tree_Node *)this,
			 (stream == stdout ? USE_STDOUT : USE_STDERR),
			 position);
}

void Virtual_Node::findAndApplyExceptionHandler (Task_Tree_Ref parentNode,
						 Task_Tree_Ref failureNode,
						 Exception_Instance *exception,
						 const void *data,
						 BOOLEAN isRefCountFailureData)
{ 
  setVirtualAddressIfNeeded(parentNode);
  setVirtualAddressIfNeeded(failureNode);
  distributedFindAndApplyExceptionHandler(this, parentNode, failureNode, 
					  exception, data,
					  isRefCountFailureData);
}

void Virtual_Node::noExceptionHandler (STRING failure) const 
{
  distributedNoExceptionHandler((Task_Tree_Node *)this, failure);
}

//
// This is complicated by the fact that ADD_EVENT_P needs to invoke the
// "validState" function, but that information is not maintained in the
// virtual node.  On the other hand, the signal has to put on the event
// queue in *this* process, since the clock associated with the process
// of the virtual node may be different.  
// Resolved by adding the event everytime, but sending the virtual node
// a message requesting it to dequeue to timed event if the ADD_EVENT_P
// condition does not hold.
//
void Virtual_Node::addTimedEventAfterSignal (MSecs waitTime,
					     const Task_Tree_Ref nodeRef, 
					     Signal_Enum signal,
					     State_Enum state)
{
  addTimedEvent(absoluteTime(waitTime), nodeRef, signal);
  if (signal == Start_Terminating_Signal) {
    setVirtualAddressIfNeeded(nodeRef);
    distributedCheckTimedEventAfterSignal(this, nodeRef, signal, state);
  }
}

/**************************************************************************
 *
 * Functions to map between a virtual address and a virtual node
 *
 * AT SOME POINT, SHOULD IMPLEMENT USING A HASH TABLE!
 *
 **************************************************************************/

/**************************************************************************
 *
 * FUNCTION: void rememberVirtual(void)
 *
 * DESCRIPTION: Memoize the node, so that it can be recalled by its 
 *              virtual address
 *
 **************************************************************************/
void Task_Tree_Node::rememberVirtual (void)
{
  virtualNodes.insertLast(this);
}

/**************************************************************************
 *
 * FUNCTION: Task_Tree_Node_Ptr recallVirtual (const Virtual_Address &vaddress)
 *
 * DESCRIPTION: Find the virtual node associated with the given
 *              virtual address
 *
 **************************************************************************/
Task_Tree_Node_Ptr
Task_Tree_Node::recallVirtual (const Virtual_Address &vaddress)
{
  Const_Task_Tree_Node_Iterator nodeIter(virtualNodes);
  Virtual_Address virtualAddress;

  for (; nodeIter; nodeIter++) {
    virtualAddress = (*nodeIter())->getVirtualAddress();
    if (vaddress.id == virtualAddress.id &&
	streq(vaddress.host, virtualAddress.host) &&
	streq(vaddress.creator, virtualAddress.creator)) {
      return (Virtual_Node_Ptr)**nodeIter();
    }
  }
  return NULL;
}

/**************************************************************************
 *
 * FUNCTION: void forgetVirtual (const Virtual_Address &vaddress) 
 *
 * DESCRIPTION: Un-memoize the node, freeing up storage
 *
 **************************************************************************/
void Task_Tree_Node::forgetVirtual (const Virtual_Address &vaddress)
{
  Task_Tree_Node_Iterator nodeIter(virtualNodes);
  Virtual_Address virtualAddress;

  for (; nodeIter; nodeIter++) {
    virtualAddress = (*nodeIter())->getVirtualAddress();
    if (vaddress.id == virtualAddress.id &&
	streq(vaddress.host, virtualAddress.host) &&
	streq(vaddress.creator, virtualAddress.creator)) {
      nodeIter.removeCurrent();
      break;
    }
  }
}

/**************************************************************************
 *
 * FUNCTION: Task_Tree_Node_Ptr getVirtual (const Virtual_Address &vaddress)
 *
 * DESCRIPTION: Find or create the virtual node associated with the
 *              given virtual address
 *
 **************************************************************************/
Task_Tree_Node_Ptr Task_Tree_Node::getVirtual (const Virtual_Address &vaddress,
					       STRING nodeName)
{
  Task_Tree_Node_Ptr virtualNode = recallVirtual(vaddress);

  return (virtualNode ? virtualNode : new Virtual_Node(nodeName, vaddress));
}


/**************************************************************************
 *
 * FUNCTION: void setVirtualAddress (STRING host, STRING creator, 
 *                                   unsigned int id)
 *
 * DESCRIPTION: Set the virtual address of the node, and memoize the node.
 *
 **************************************************************************/
void  Task_Tree_Node::setVirtualAddress (STRING host, STRING creator, 
					 unsigned int id)
{
  if (host == NULL || strlen(host) == 0 || 
      creator == NULL || strlen(creator) == 0) {
    tcmError("setVirtualAddress: Null host and/or creator for %s\n",
	     instanceName());
  } else if (validVirtualAddress()) {
    tcmError("setVirtualAddress: Address for %s already set\n",
	     instanceName());
  } else {
    _vaddress.host = strdup(host);
    _vaddress.creator = strdup(creator);
    _vaddress.id = id;
    this->id = id;
    rememberVirtual();
  }
}
